from sentence_transformers.losses import BatchHardTripletLossDistanceFunction
import torch


class MultiDimensionalTripletLoss():
    def __init__(self, triplet_margin=1, dynamic_margin=True, discrete_labels=True):
        self.triplet_margin = triplet_margin
        self.dynamic_margin = True
        self.discrete_labels = discrete_labels
    
    def __call__(self, embeddings, labels):
        distance_metric = BatchHardTripletLossDistanceFunction.eucledian_distance
        emb_dist_diff = self.pairwise_dist_diff(distance_metric(embeddings))

        if self.discrete_labels:
            label_dist_diff = self.pairwise_dist_diff(self.label_pairwise_dist(labels))
        else:
            label_dist_diff = self.pairwise_dist_diff(distance_metric(labels))

        if self.dynamic_margin:
            triplet_margin = -1 * label_dist_diff * self.triplet_margin
        else:
            triplet_margin = self.triplet_margin

        triplet_loss = emb_dist_diff + triplet_margin
        mask = self.get_triplet_mask(label_dist_diff)
        triplet_loss = mask.float() * triplet_loss

        triplet_loss = triplet_loss.reshape(triplet_loss.shape[0], -1)
        triplet_loss = triplet_loss.max(1).values
        triplet_loss = triplet_loss.mean()
        return triplet_loss


    
    @staticmethod
    def pairwise_dist_diff(pairwise_dist):
        positive_dist = pairwise_dist.unsqueeze(2)
        negative_dist = pairwise_dist.unsqueeze(1)
        return positive_dist - negative_dist
    
    @staticmethod
    def label_pairwise_dist(labels):
        return 1 - (labels.unsqueeze(0) * labels.unsqueeze(1)).sum(2) / (labels.sum(1) + 1e-16)

    @staticmethod
    def get_triplet_mask(label_dist_diff):
        indices_equal = torch.eye(label_dist_diff.size(0), device=label_dist_diff.device).bool()
        indices_not_equal = ~indices_equal
        i_not_equal_j = indices_not_equal.unsqueeze(2)
        i_not_equal_k = indices_not_equal.unsqueeze(1)
        j_not_equal_k = indices_not_equal.unsqueeze(0)
        distinct_indices = (i_not_equal_j & i_not_equal_k) & j_not_equal_k
        valid_labels = label_dist_diff < 0
        return valid_labels & distinct_indices


